
import deepl
# from API_KEYS import DEEPL # import deepl password

import time


def translate(text, engine="google", sleep=0):
    import translators as ts  # this would automatically show server when import

    # baidu alibaba caiyun is very slow
    if engine == "google":
        tr = ts.google(text, from_language='zh', to_language='en')
    elif engine == "caiyun":
        tr = ts.caiyun(text, from_language='zh', to_language='en')
    time.sleep(sleep)
    return tr


def translate_deepl(text, password):
    translator = deepl.Translator(password)
    tr = translator.translate_text(text, source_lang="ZH", target_lang="EN-US").text
    return tr



import pandas as pd
from tqdm import tqdm


def chunk_apply(df, func, args=(), n=10_000):

    m = df.shape[0]//n + 1
    list_dfjb = (df[i:i+n] for i in range(0, df.shape[0], n))

    listr_res = []

    for i in tqdm(list_dfjb, total=m):
        _res = i.pipe(func, *args)
        listr_res.append(_res)

    res = pd.concat(listr_res, axis=0)
    return res




"""
Useful plots in data anlysis
"""
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns

if __name__ != "__main__":
    __version__ = 0.1


def plot_scatter_with_text(
        data, x, y, z, hue=None, style=None,
        text_condition=None, text_adjust=False, adjust_precision=0.1):
    """
    Use seaborn, matplotlib and adjustText to plot scatter with text
    """
    data = data.copy().reset_index(drop=True)
    ax = sns.scatterplot(data=data, x=x, y=y, hue=hue, style=style)
    if isinstance(text_condition, pd.Series):
        data = data[text_condition].reset_index(drop=True)
    texts = [ax.annotate(text, (data.loc[i, x], data.loc[i, y])) for i, text in enumerate(data[z].values)]
    if text_adjust:
        from adjustText import adjust_text
        adjust_text(texts, precision=adjust_precision)
    return ax


def plot3D(x_min, x_max, y_min, y_max, x_n, y_n, z_func,
           x_label='x', y_label='y', z_label='z', zlim=None,
           figsize=(8, 6), cmap='summer', alpha=0.8, rotate=225):
    """
    This code is used to plot the 3D figure of a function with two variables
    """
    x_vals = np.linspace(x_min, x_max, x_n)
    y_vals = np.linspace(y_min, y_max, y_n)
    x_mesh, y_mesh = np.meshgrid(x_vals, y_vals, indexing='ij')
    try:
        z_vals = z_func(x_mesh, y_mesh)
    except ValueError:
        z_vals = np.zeros((x_n, y_n))
        for i in range(x_n):
            for j in range(y_n):
                z_vals[i, j] = z_func(x_vals[i], y_vals[j])
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111, projection='3d')
    ax = fig.gca(projection='3d')
    ax.plot_surface(x_mesh, y_mesh, z_vals, cmap=cmap, alpha=alpha)  # winter
    ax.set(xlabel=x_label, ylabel=y_label, zlabel=z_label, zlim=zlim,)
    ax.zaxis.set_rotate_label(False)  # ?
    ax.view_init(ax.elev, rotate)  # view direction
    plt.close()
    return fig



"""
This code is generally borrown from under-maintained pkg `japandas`
and used to turn 「全角」strings to 「半角」strings.
"""

__version__ = 0.1

from unicodedata import normalize

# soundmarks require special handlings
_ZALPHA = ('ＡＢＣＤＥＦＧＨＩＪＫＬＭＮＯＰＱＲＳＴＵＶＷＸＹＺ'
           'ａｂｃｄｅｆｇｈｉｊｋｌｍｎｏｐｑｒｓｔｕｖｗｘｙｚ')
_ZSYMBOL = '！＂＃＄％＆＇（）＊＋，－．／：；＜＝＞？＠［＼］＾＿｀｛｜｝～　'
_ZDIGIT = '０１２３４５６７８９'

# mapping from full-width to half-width
_ALPHA_MAPPER = {c: normalize('NFKC', c) for c in _ZALPHA}
_DIGIT_MAPPER = {c: normalize('NFKC', c) for c in _ZDIGIT}
_SYMBOL_MAPPER = {c: normalize('NFKC', c) for c in _ZSYMBOL}

# adding symbols that un-normalizable
# https://www.utf8-chartable.de/unicode-utf8-table.pl?start=12224&names=-&utf8=string-literal
_ZSYMBOL_MAPPER = {"〜": "~", }
_SYMBOL_MAPPER.update(_ZSYMBOL_MAPPER)


def _ord_dict(dict):
    return {ord(k): v for k, v in dict.items()}


# for unicode.translate
_Z2H_ALPHA = _ord_dict(_ALPHA_MAPPER)
_Z2H_DIGIT = _ord_dict(_DIGIT_MAPPER)
_Z2H_SYMBOL = _ord_dict(_SYMBOL_MAPPER)

mapper = dict()
mapper.update(_Z2H_ALPHA)
mapper.update(_Z2H_DIGIT)
mapper.update(_Z2H_SYMBOL)


def str_z2h(string: str, ):
    try:
        res = string.translate(mapper)
        return res
    except AttributeError:
        return string



"""
If you use mac, this simple code can help to inform you when running a time-consuming program.
"""
import os


def say(word, repeat=1, voice="kyoko"):
    os.system(f"say -v {voice} {word*repeat}")


def owari_voice(repeat=2, error=False, jp_voice=True):
    if jp_voice:
        word_fin = "コード実行が終わりました。"
        word_err = "エラーが発生しました。"
        voice_fin = 'kyoko'
        voice_err = 'kyoko'
    else:
        word_fin = "Code have been finished."
        word_err = "Bug have been occured."
        voice_fin = 'Good\\ News'
        voice_err = 'Bad\\ News'

    if not error:
        say(word_fin, repeat, voice_fin)
    else:
        say(word_err, repeat, voice_err)


def owari(nys=True, repeat=2, **args):

    nys_path = "../asset/nps.wav"

    if nys & os.path.exists(nys_path):
        for _ in range(repeat):
            os.system(f"afplay {nys_path}")
    else:
        owari_voice(repeat=repeat, **args)



"""
General tools used to draw figures in data analysis
"""

import numpy as np
import matplotlib as mpl
from cycler import cycler
import matplotlib.pyplot as plt
import seaborn as sns



def set_style(color_style=None, color_order=None,
              tick_direction="in", ticksize=3, tickwidth=0.7,
              half_spine=False, full_tick=True, full_axis_label=False,
              grid_axis="both", grid_which='major', grid_alpha=0.2, grid_color='b',
              marker_cycle=False
              ):
    # style
    sns.set_style("ticks")  # "white", "whitegrid", "ticks"
    sns.set_context("paper")

    # font
    nice_fonts = {
        "font.family": "sans-serif",
        # "font.family": "serif",
        "font.serif": ["Palatino", "Adobe Garamond Pro"],
        "font.sans-serif": ["Assistant", "Helvetica", "Arial"],  #
        # "text.usetex": True,
        'pdf.fonttype': 42,  # default Type 3, does not support for some fonts
    }
    mpl.rcParams.update(nice_fonts)

    # font size
    nice_fonts = {
        "font.size": 10,
        "axes.titlesize": 11,
        "axes.labelsize": 10,
        "legend.fontsize": 10,
        "xtick.labelsize": 10,
        "ytick.labelsize": 10,
    }
    mpl.rcParams.update(nice_fonts)

    # lines
    nice_lines = {
        "axes.linewidth": 0.7,
        "lines.linewidth": 1.2,
        "lines.markersize": 4,
        "grid.linewidth": 0.7
    }
    mpl.rcParams.update(nice_lines)

    # ticks
    plt.rc('xtick', direction=tick_direction)
    plt.rc('xtick.major', size=ticksize, width=tickwidth)
    plt.rc('ytick', direction=tick_direction)
    plt.rc('ytick.major', size=ticksize, width=tickwidth)
    if full_tick:
        plt.rcParams['xtick.top'] = True
        plt.rcParams['ytick.right'] = True
    if full_axis_label:
        plt.rcParams['xtick.labeltop'] = True
        plt.rcParams['ytick.labelright'] = True

    # spine
    if half_spine:
        plt.rc('axes.spines', top=False, right=False)

    # grid
    plt.rcParams['axes.grid'] = True
    plt.rc('axes.grid', axis=grid_axis, which=grid_which)
    plt.rc('grid', color=grid_color, alpha=grid_alpha)

    # color
    color_cycle = cycler('color', get_colors(color_style, color_order))
    plt.rcParams['axes.prop_cycle'] = color_cycle  # plt.cycler(color=get_colors(color_order))

    # marker
    if marker_cycle:
        # can use plt.cycler then no need to import cycler
        # plt.cycler(color=get_colors(color_order)), marker=get_markerstyle())
        plt.rcParams['axes.prop_cycle'] = (color_cycle + cycler('marker', get_markerstyle()))

    # layout
    # plt.rcParams['figure.constrained_layout.use'] = True

    # latex
    # nice_latex = {
    # "pgf.texsystem": "pdflatex",
    # }
    # mpl.rcParams.update(nice_latex)


def get_colors(style=None, order=None, mono=False):
    if style is None:
        colors = np.array([
            "#eb4d55",  # red
            # '#E24A33',  # red
            "#5f6caf",  # blue
            # '#348ABD',  # blue
            "#32afa9",  # green
            # '#8EBA42',  # green
            '#FBC15E',  # yellow
            # "#ffd369",  # yellow
            '#FFB5B8',  # pink
            # '#988ED5',  # purple
            "#916dd5",  # purple
            '#777777',  # grey
            # "#333333",  # black
        ])
    elif style == "ggplot":
        colors = ['#E24A33', '#348ABD', '#988ED5', '#777777', '#FBC15E', '#8EBA42', '#FFB5B8']
    elif style == "default":
        colors = ['#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd',
                  '#8c564b', '#e377c2', '#7f7f7f', '#bcbd22', '#17becf']
    else:
        raise ValueError("color style only support None or ggplot or default")
    if order:
        colors = [colors[i] for i in order]
    if mono:
        colors = ["black"]
    return colors


def get_greys(n, reverse=True, start=False, ):
    greys = [str(g) for g in np.linspace(0.1, 0.9, n)]
    if reverse:
        greys.reverse()
    if start:
        greys = [str(g) for g in np.linspace(0.1, 0.9, n-1)]
        greys = get_colors()[0] + greys
    return greys


def get_linestyle(dot_line=False, marker=True, same_marker=False, same_marker_type=0):
    if dot_line:
        linestyle = ['-', '--', '-.', ':', ]
    else:
        linestyle = ['-'] * 4
    if marker:
        if same_marker:
            markerstyle = get_markerstyle()[same_marker_type:same_marker_type+1] * 4
        else:
            markerstyle = get_markerstyle()[:len(linestyle)]
        linestyle = [l+m for l, m in zip(linestyle, markerstyle)]
    return linestyle


def get_markerstyle():
    markers = ['o', '^', 's', 'd', 'x', '+', '*']
    return markers


def set_axes(axes, half_spine=False, full_tick=True, xgrid=True, grid_alpha=0.2, grid_color='b'):
    # all can be set in set_style, here only leave the option
    for ax in axes:
        if half_spine:
            ax.spines['right'].set_visible(False)
            ax.spines['top'].set_visible(False)
        if full_tick:
            ax.tick_params(top=True, right=True)
        ax.grid(axis='y', color=grid_color, alpha=grid_alpha,)
        if xgrid:
            ax.grid(axis='x', color=grid_color, alpha=grid_alpha,)


def add_super_label_for_multi_axes(fig, ylabel=None, xlabel=None):
    ax = fig.add_subplot(111, frameon=False)
    ax.tick_params(labelcolor='none', top=False, bottom=False, left=False, right=False)
    ax.grid(False)
    ax.set_ylabel(ylabel,)
    ax.set_xlabel(xlabel)


def set_size(fig, fraction=0.7, row=1, h_pad=None, w_pad=None, rect=None):
    """
    fraction: fraction of textwidth
    row: level of height, when 1 perfect for 1x1 figure, for other shapes, need some arts
    """
    width = 469.75502  # wp \the\textwidth in latex
    fig_width_pt = width * fraction
    inches_per_pt = 1 / 72.27
    golden_ratio = (5**.5 - 1) / 2  # 0.618 = 1 / 1.618
    fig_width_in = fig_width_pt * inches_per_pt
    fig_height_in = fig_width_in * golden_ratio * row  # (shape[0]/shape[1])

    fig.set_size_inches(fig_width_in, fig_height_in)
    fig.tight_layout(h_pad=h_pad, w_pad=w_pad, rect=rect)


def save_fig(fig, name, output_path="../Figures/"):
    # alternatively saving in eps
    # transparent=Ture
    # pad_inches=0
    fig.savefig(output_path+name, format='pdf', bbox_inches='tight')



"""
Useful plots in data anlysis
"""
import altair as alt

if __name__ != "__main__":
    __version__ = 0.1

def plot_interactive_lines(data, x, y, hue, alpha=0.2, point_size=15, line_size=1.5):
    """
    Use altair to plot multiple lines with interactive selection
    """
    highlight = alt.selection(type='single', fields=[hue], on='mouseover', nearest=True, bind='legend')
    selection = alt.selection_multi(fields=[hue], on='mouseover', bind='legend')
    base = alt.Chart(data).encode(
        x=x,
        y=y,
        color=hue,
        tooltip=[hue],
    )
    points = base.mark_circle().encode(
        opacity=alt.condition(selection, alt.value(1), alt.value(alpha)),
        size=alt.condition(~highlight, alt.value(point_size), alt.value(point_size*2))
    ).add_selection(highlight)
    lines = base.mark_line().encode(
        opacity=alt.condition(selection, alt.value(1), alt.value(alpha)),
        size=alt.condition(~highlight, alt.value(line_size), alt.value(line_size*2))
    ).add_selection(selection)
    return lines + points




import pickle
from urllib.request import urlopen
# import dill as pickle  # use dill to help io model instance


def write_file(file, file_path):
    with open(file_path, 'wb') as f:
        pickle.dump(file, f)


def read_file(file_path):
    if file_path.startswith('http'):
        with urlopen(file_path,) as f:
            file = pickle.load(f)
    else:
        with open(file_path, 'rb') as f:
            file = pickle.load(f)
    return file



"""
This code turn country names to continents.
"""

__version__ = 0.1

import country_converter as coco
from pycountry_convert import country_alpha2_to_continent_code, country_name_to_country_alpha2
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)

Continents = {
    'NA': 'North America',
    'SA': 'South America',
    'EU': 'Europe',
    'AS': 'Asia',
    'OC': 'Australia',
    'AF': 'Africa',
}


def country_to_continent(countries):
    """
    Convert countries to continents.
    A single country is ok but must be in a list.
    """
    continents = []
    for country in countries:
        try:
            iso = country_name_to_country_alpha2(country)
        except Exception:
            iso = coco.convert(names=country, to='ISO2')
        continent = country_alpha2_to_continent_code(iso)
        continent = Continents[continent]
        continents.append(continent)
    return continents



"""
Send an email from your G-mail account.
ref: https://realpython.com/python-send-email/#sending-your-plain-text-email
"""

__version__ = 0.1

import smtplib
import ssl
# from API_KEYS import gmail_alalaki as mail_password  # Import your password
from email.mime.text import MIMEText


def sendmail(
    from_email="alalaki.koyomi@gmail.com",  # Enter your address
    to_email="akaguro.koyomi@gmail.com",  # Enter receiver address, can be a list
    subject="Test Test",
    body=""",
Hi,

This message is sent from Python.
""",
    mail_password=None
):
    port = 465  # For SSL
    smtp_server = "smtp.gmail.com"
    password = mail_password
    message = MIMEText(body)
    message['Subject'] = subject
    message['From'] = from_email
    message['To'] = ", ".join(to_email) if isinstance(to_email, list) else to_email

    context = ssl.create_default_context()
    with smtplib.SMTP_SSL(smtp_server, port, context=context) as server:
        server.login(from_email, password)
        server.sendmail(from_email, to_email, message.as_string())


# if __name__ == "__main__":
#     sendmail()



"""
Factory for data cleaning functions
Under progress
"""

__version__ = 0.0

def clean_column_name(df, n):
    """
    combine the top n rows to be the column name
    """
    column = df.iloc[:n, :]
    column = column.fillna('').apply(lambda x: '_'.join(
        x.astype(str).drop_duplicates().values))
    df = df.iloc[n:, :].reset_index(drop=True)
    df.columns = column
    return df



from .ipy_print import prints
from .print import printt, printf
from .sendmail import sendmail
from .owari import owari
# from .wsummary import wsummary
# from .z2h import str_z2h
# from .plot3D import plot3D
# from .cleandata import clean_column_name
# from .download import download_file
# from .country2continent import country_to_continent


__all__ = [
    "prints", "printt", "printf", "sendmail", "owari",
    # "wsummary",
    # "str_z2h", "plot3D",
    # "clean_column_name",
    # "download_file",
]



"""
This code provide weighted statistic summary for data
Source: stolen from Spencer Lyon's "Replicating ADH 2013"
"""

import numpy as np
import pandas as pd

# Helper routines for weighted descriptive stats
def xbar(x, w):
    return np.dot(x, w) / np.sum(w)


def w_stdev(x, w):
    n = x.shape[0]
    W = np.sum(w)
    xb = xbar(x, w)
    return np.sqrt((n/(W * (n - 1))) * np.sum(w * (x - xb)**2))


def w_mean_std(x, w):
    return xbar(x, w), w_stdev(x, w)


def wquantile(vals_in, weights_in, probs_in):
    """
    Taken from StatsBase.jl
    """
    weights = np.asarray(weights_in)
    probs = np.asarray(probs_in)

    if isinstance(vals_in, pd.DataFrame):
        datas = [wquantile(vals_in[c], weights, probs) for c in vals_in.columns]
        return pd.concat(datas, axis=1)

    vals = np.asarray(vals_in)

    # full sort
    inds = np.argsort(vals)
    v = vals.copy()[inds]
    w = weights.copy()[inds]
    wsum = w.sum()

    # prepare percentiles
    p_inds = np.argsort(probs)
    p = np.clip(probs.copy()[p_inds], 1e-10, 1-1e-10)

    # prepare out vector
    N = v.size
    out = np.empty(p.size)
    out[:] = v[-1]

    # start looping on quantiles
    cumulative_weight, Sk, Skold = 0.0, 0.0, 0.0
    vk, vkold = 0.0, 0.0
    k = 0
    for i in range(p.size):
        h = p[i] * (N - 1) * wsum
        if h == 0:
            # happens when N or p or wsum equal zero
            out[p_inds[i]] = v[0]
        else:
            while Sk <= h:
                # happens in particular when k == 0
                vk = v[k]
                wk = w[k]
                cumulative_weight += wk
                if k >= N:
                    # out was initialized with maximum v
                    return out
                k += 1
                Skold, vkold = Sk, vk
                vk = v[k]
                wk = w[k]
                Sk = (k - 1) * wk + (N - 1) * cumulative_weight
            # in particular, Sk is different from Skold
            g = (h - Skold) / (Sk - Skold)
            out[p_inds[i]] = vkold + g * (vk - vkold)

    if isinstance(vals_in, pd.Series):
        res = pd.DataFrame(index=probs, data={vals_in.name: out})

    res.index.name = "quantile"
    return res


def wsummary(df, weights, qs=None):
    "looks like Stata's summarize variable [aw=weights], detail"
    if isinstance(df, pd.Series):
        return wsummary(df.to_frame(), weights, qs)

    if not isinstance(df, pd.DataFrame):
        msg = "wsummary only works with pandas Series or DataFrame"
        raise ValueError(msg)

    if qs is None:
        qs = [0.01, 0.05, 0.1, 0.25, 0.5, 0.75, 0.9, 0.95, 0.99]

    if len(qs) > 0:
        quants = wquantile(df, weights, qs)
        quants.index.name = None
        new_index = ["{0:d}%".format(int(i*100)) for i in quants.index]
        quants.index = ["{0:d}%".format(int(i*100)) for i in quants.index]
        quantsT = quants.T
    else:
        quantsT = pd.DataFrame(index=df.columns)

    # add other descriptive stats
    quantsT["count"] = df.count()
    quantsT["mean"] = df.aggregate(lambda x: xbar(x, weights))
    quantsT["std"] = df.aggregate(lambda x: w_stdev(x, weights))
    quantsT["min"] = df.min()
    quantsT["max"] = df.max()
    out = quantsT.T

    if len(qs) > 0:
        row_order = ["count", "mean", "std", "min"] + new_index + ["max"]
        return out.loc[row_order]
    else:
        return out



"""
This code is used to add paths to system path so that ipynb can import python files in other folders
"""

import sys
import os


def addpath(path="Code"):
    "default path at parent directory"
    module_path = os.path.abspath(os.path.join(os.pardir, path))
    if module_path not in sys.path:
        sys.path.append(module_path)




import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from sklearn.metrics import confusion_matrix, accuracy_score, balanced_accuracy_score, classification_report, r2_score, mean_squared_error

from mpl_set import set_style, set_size, save_fig


def scoring(y_true, y_pred, y_score=None, target_names=None):
    _as = accuracy_score(y_true, y_pred)
    print(f"accuracy_score: {_as}")
    bas = balanced_accuracy_score(y_true, y_pred)
    print(f"balanced_accuracy_score: {bas}")
    # ras = roc_auc_score(y_true, y_score)
    # print(f"roc_auc_score: {ras}")
    cr = classification_report(y_true, y_pred, target_names=target_names)
    print(f"classification_report: {cr}")


def plot_confusion(y_true, y_predict, ticklabels, normalize="true", only_wrong=False,
                   figsize=(22, 22), fmt='.4f', major_lines=None,
                   save_path=None
                   ):
    confusion = confusion_matrix(y_true, y_predict, normalize=normalize)
    if only_wrong:
        np.fill_diagonalgonal(confusion, 0)

    fig, ax = plt.subplots(figsize=figsize)
    sns.heatmap(confusion.T, ax=ax, square=True, cbar=False, cmap='rocket_r', annot=True,
                fmt=fmt,  # if count, fmt='d', if normalzie fmt='.2f'
                xticklabels=ticklabels, yticklabels=ticklabels)
    if major_lines is not None:
        ax.hlines(major_lines, *ax.get_xlim(), color="C1")
        ax.vlines(major_lines, *ax.get_xlim(), color="C1")
    plt.xlabel('true label')
    plt.ylabel('predicted label')

    if save_path is not None:
        save_fig(fig, save_path)


def bic_scorer(estimator, X, y):
    # ref: https://github.com/scikit-learn/scikit-learn/blob/844b4be24/sklearn/linear_model/_least_angle.py#L1952
    n = len(y)
    y_pred = estimator.predict(X)
    mse = np.average((y - y_pred) ** 2, axis=0)  # (y - Xβ) / N
    sigma2 = np.var(y)
    eps64 = np.finfo("float64").eps
    K = np.log(n)
    mask = np.abs(estimator.coef_) > np.finfo(estimator.coef_.dtype).eps
    df = np.sum(mask)

    score = (n * mse / (sigma2 + eps64) + K * df)
    return score


def bic_scorer_neg(estimator, X, y):
    return bic_scorer(estimator, X, y) * -1


def gic_scorer(estimator, X, y, K="lnlp"):
    # exactly the same as bic except K
    n = len(y)
    p = X.shape[1]
    y_pred = estimator.predict(X)
    mse = np.average((y - y_pred) ** 2, axis=0)
    sigma2 = np.var(y)
    eps64 = np.finfo("float64").eps
    if K == "lnlp":
        K = np.log(n) * np.log(p)
    elif K == "llnlp":
        K = np.log(np.log(n)) * np.log(p)
    elif K == "2ln":
        K = 2*np.log(n)
    else:
        raise ValueError("K should be lnlp or llnlp or 2ln")
    mask = np.abs(estimator.coef_) > np.finfo(estimator.coef_.dtype).eps
    df = np.sum(mask)

    score = (n * mse / (sigma2 + eps64) + K * df)
    return score


def best_alpha_index(results):
    # ref: https://github.com/scikit-learn-contrib/lightning/issues/84
    # ref: https://scikit-learn.org/stable/modules/model_evaluation.html#implementing-your-own-scoring-object
    K = len([x for x in list(results.keys()) if x.startswith('split')])
    alpha_range = results['param_ridge__alpha'].data

    mean_per_alpha = pd.Series(results['mean_test_score'], index=alpha_range)
    std_per_alpha = pd.Series(results['std_test_score'], index=alpha_range)
    sem_per_alpha = std_per_alpha / np.sqrt(K)

    max_score = mean_per_alpha.max()
    sem = sem_per_alpha[mean_per_alpha.idxmax()]
    best_alpha = mean_per_alpha[mean_per_alpha >= max_score - sem].index.max()

    best_alpha_index = int(np.argwhere(alpha_range == best_alpha)[0])

    return best_alpha_index


def plot_model_fit(y_test, y_test_predict, lim=(7, 12), figsize=(8, 5), label="Log Wage", save_path=None):
    r2 = r2_score(y_test, y_test_predict)
    string_score = f"R2: {r2:.3f}"
    mse = mean_squared_error(y_test, y_test_predict)
    string_score += f"\nMSE: {mse:.3f}"
    set_style(color_style="default")
    fig, ax = plt.subplots(figsize=figsize)
    # ax.scatter(y_test, y_test_predict, alpha=0.02, c="C0")
    ax.plot(y_test, y_test_predict, 'o', alpha=0.1, c="C0", rasterized=True)  # this reduce figure size
    ax.plot([0, 1], [0, 1], transform=ax.transAxes, ls="--", c="C1")
    ax.text(lim[0]*1.1, lim[1]*0.9, string_score)
    # plt.title(f"Lasso, α={model.alpha_:.4}")
    plt.ylabel(label+", Model Predictions")
    plt.xlabel(label+", Truths")
    plt.xlim(lim)
    plt.ylim(lim)

    if save_path is not None:
        save_fig(fig, save_path)



"""
This code is used to download files
"""

import os
import cgi
import requests
import shutil

if __name__ != "__main__":
    __version__ = 0.1


def download_file(url, path, file_name=None):
    """
    Download file from url to path.
    If file_name is not given, it would either look for the filename
    in Content-Disposition header or use the filename in url.
    """
    res = requests.get(url, stream=True)
    if res.status_code != 200:
        raise ValueError(f'Failed to download: {res.status_code}')

    if file_name is None:
        if 'Content-Disposition' in res.headers:
            params = cgi.parse_header(
                res.headers.get('Content-Disposition', ''))[-1]
            if 'filename' not in params:
                raise ValueError('Could not find a filename')
            file_name = params['filename']
        else:
            file_name = os.path.basename(url)

    abs_path = os.path.join(path, os.path.basename(file_name))
    with open(abs_path, 'wb') as target:
        res.raw.decode_content = True
        shutil.copyfileobj(res.raw, target)

    print(f"Download {file_name}")



"""
This code is used to print a pandas dataframe with customized length of rows and columns.
Ref: https://pandas.pydata.org/pandas-docs/stable/user_guide/options.html

Note:
  - When setting as `exec_files` in `ipython_config`, the `prints` function `display` the DataFrame.
  - Alternatively when import the file, the `prints` function `print` the DataFrame.
"""

import pandas as pd


if __name__ != "__main__":
    __version__ = 0.1


def prints(x, r=None, c=None, w=-1):
    """Display a DataFrame in Ipython by specified Column numbers and Row numbers.

    Parameters
    ----------
    x : pd.DataFrame or pd.Series

    r : int, optional
        No. of rows shown. By default None, display all rows
    c : int, optional
        No. of columns shown. By default None, display all columns,
    w : int, optional
        Max column width. By default -1.
    """
    # m = pd.options.display.max_rows
    # n = pd.options.display.max_columns
    # w = pd.options.display.max_colwidth

    if isinstance(x, pd.DataFrame):

        with pd.option_context('display.min_rows', r, 'display.max_rows', r, 'display.max_columns', c, 'display.max_colwidth', w):
            if __name__ == "__main__":
                display(x)
            else:
                print(x)

    elif isinstance(x, pd.Series):
        with pd.option_context('display.min_rows', r, 'display.max_rows', r, 'display.max_colwidth', w):
            if __name__ == "__main__":
                display(x)
            else:
                print(x)

    else:
        raise ValueError('Input must be a pandas object.')




import pandas as pd


def add_share_columns(df, base_line_idx, column_names=["Value", "Share"], formatting=True):
    dt = pd.DataFrame()
    for i, c in df.items():
        pct = c / c[base_line_idx]
        if formatting:
            pct = pct.map(lambda x: f'{abs(x):.3f}'.lstrip('0'))
        pct[base_line_idx] = "-"

        comb = pd.DataFrame({column_names[0]: c, column_names[1]: pct})
        comb.columns = [[i, i], comb.columns]
        dt = pd.concat([dt, comb], axis=1)
    return dt


def add_share_rows(df, base_line_idx, exp_lines_idx=None, row_name=" ", formatting=True):
    dt = pd.DataFrame()
    for i, r in df.iterrows():
        dt = dt.append(r)
        if i not in exp_lines_idx:
            pct = r/df.iloc[base_line_idx, :]
            if formatting:
                # pct = pct.map(lambda x: "("+str(int(x*100))+"%)")
                pct = pct.map(lambda x: "("+f'{abs(x):.2f}'.lstrip('0')+")")
            dt = dt.append(pct.rename(row_name))
    return dt



from datetime import datetime


def printt(x):
    print(datetime.now().strftime('%Y-%m-%d %H:%M:%S'), x)


def printf(x, f=4):
    # print([f'{i:.{f}f}' for i in x])
    print([round(i, f) for i in x])



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['autil']

package_data = \
{'': ['*']}

install_requires = \
['adjustText>=0.7.3,<0.8.0',
 'altair>=4.1,<5.0',
 'country_converter>=0.7.3,<0.8.0',
 'cycler>=0.10,<0.11',
 'deepl>=1.6.0,<2.0.0',
 'linearmodels>=4.24,<5.0',
 'matplotlib>=3.4,<4.0',
 'numpy>=1.21,<2.0',
 'pandas>=1.3,<2.0',
 'pycountry_convert>=0.7.2,<0.8.0',
 'requests>=2.26,<3.0',
 'scikit_learn>=1.0,<2.0',
 'seaborn>=0.11,<0.12',
 'tqdm>=4.62,<5.0',
 'translators>=4.9.5,<5.0.0']

setup_kwargs = {
    'name': 'autil',
    'version': '0.1.0',
    'description': 'Some Python snippets for daily use',
    'long_description': None,
    'author': 'alalalalaki',
    'author_email': 'harlan.zhu@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)




import re


def remove_reg_var(file_path: str, remove_var: list = ["year"], next_line=True) -> None:
    """
    Remove certain variables shown in latex table
    `remove_var` takes a list of the start words of the removed varaibles (case sensitive)
    """

    with open(file_path, 'r', encoding='utf8') as f:
        lines = f.readlines()

    remove_idx = []
    for i, li in enumerate(lines):
        for s in remove_var:
            if li.startswith(s):
                remove_idx.append(i)
                if next_line:
                    remove_idx.append(i + 1)

    lines = [li for i, li in enumerate(lines) if i not in remove_idx]

    with open(file_path, 'w', encoding='utf8') as f:
        f.writelines(lines)


def remove_preceding_zero(file_path: str) -> None:

    with open(file_path, 'r', encoding='utf8') as f:
        lines = f.readlines()

    lines = [re.sub('(?<!\d)0\.', '.', li) for li in lines]

    with open(file_path, 'w', encoding='utf8') as f:
        f.writelines(lines)


def remove_float_zero(file_path: str, lines_idx: list) -> None:
    """
    Remove .00... for int rows
    However this will remove also not float rows with some entry happen to be .00
    Thus lines_idx is must
    """

    with open(file_path, 'r', encoding='utf8') as f:
        lines = f.readlines()

    lines = [re.sub('(?<=\d)\.0+\s', ' ', li) if idx in lines_idx else li for idx, li in enumerate(lines)]

    with open(file_path, 'w', encoding='utf8') as f:
        f.writelines(lines)


def reduce_float_digit(file_path: str, digit: int, lines_idx: list) -> None:
    """
    Reduce float digits for specific lines
    """

    with open(file_path, 'r', encoding='utf8') as f:
        lines = f.readlines()

    def reduce_digit(x): return f'{float(x.group(0)):.{digit}f}'.lstrip('0')
    lines = [re.sub('\.\d+', reduce_digit, li) if idx in lines_idx else li for idx, li in enumerate(lines)]

    with open(file_path, 'w', encoding='utf8') as f:
        f.writelines(lines)


def replace_string(file_path: str, pat: str, repl: str, lines_idx: list = None) -> None:
    """
    Note: to input a real \ in `repl` need \\\\
    """

    with open(file_path, 'r', encoding='utf8') as f:
        lines = f.readlines()

    if lines_idx is not None:
        lines = [re.sub(pat, repl, li) if idx in lines_idx else li for idx, li in enumerate(lines)]
    else:
        lines = [re.sub(pat, repl, li) for li in lines]

    with open(file_path, 'w', encoding='utf8') as f:
        f.writelines(lines)


def add_line(file_path: str, new_line: str, line_idx: int, linebreak: str = "\\\\ ") -> None:

    with open(file_path, 'r', encoding='utf8') as f:
        lines = f.readlines()

    lines.insert(line_idx, " " + new_line + linebreak)

    with open(file_path, 'w', encoding='utf8') as f:
        f.writelines(lines)



"""
This code is used to display two or more pandas dataframes simulatenously.
ref:
https://stackoverflow.com/questions/38783027/jupyter-notebook-display-two-pandas-tables-side-by-side
https://jakevdp.github.io/PythonDataScienceHandbook/03.06-concat-and-append.html

Update: a more easier way is to use `df.compare()`
Update: need to modify the function names and documents
"""

import pandas as pd


class display_dfs_all:
    """Display HTML representation of multiple DataFrames

    ----------
    display_dfs(df1, df2, pd.concat([df1, df2]), ...)

    # Display all columns and rows. Don't use for large data.
    ----------
    """

    def __init__(self, *args):
        if len(args) < 2:
            raise ValueError('There must be at least 2 DataFrames.')
        for i in args:
            if not isinstance(i, pd.DataFrame):
                raise ValueError('Inputs must be DataFrames.')
        self.args = args

    def _repr_html_(self):
        return ''.join([df.to_html() for df in self.args]).replace('table', 'table style="display:inline"')


class display_dfs_all_:
    """ Display HTML representation of multiple DataFrames with or without name and index

    Parameters:
    ----------
    dfs: List of DataFrames. e.g. [df1, df2, pd.concat([df1, df2]), ...]
    names: List of the name of DataFrames
    index: If show index of DataFrames. True or False

    # Can add more df.to_html(*args) like index if necessary
    # Display all columns and rows. Don't use for large data.
    ----------
   """

    def __init__(self, dfs, names=None, index=True):
        if not isinstance(dfs, list):
            raise ValueError('Input must be a list of DataFrames.')
        self.dfs = dfs
        self.names = names
        self.index = index

    def _repr_html_(self):
        # style = """<div style="float: left; padding: 5px;"><p style='font-family:"Courier New", Courier, monospace'>"""
        if self.names:
            if len(self.names) != len(self.dfs):
                raise ValueError(
                    f'Shape of names {len(self.names)} != Shape of dfs {len(self.dfs)}')
            else:
                html_str = ''.join([df.to_html(index=self.index).replace('<thead>', f'<caption>{name}</caption><thead>')
                                    for df, name in zip(self.dfs, self.names)])\
                    .replace('table', 'table style="display:inline"')  # <caption>{style}{name}</caption><thead>
        else:
            html_str = ''.join([df.to_html(index=self.index) for df in self.dfs]
                               ).replace('table', 'table style="display:inline"')
        return html_str


class display_dfs(object):
    """Display HTML representation of multiple DataFrames and names but truncated
    ----------
    display_dfns('df1', 'df2', 'pd.concat([df1, df2])', ...)

    # Name is df names
    # No. of columns and rows displayed followed defualt setting
    ----------
    """
    template = """<div style="float: left; padding: 10px;">
    <p style='font-family:"Courier New", Courier, monospace'>{0}</p>{1}
    </div>"""

    def __init__(self, *args):
        if len(args) < 2:
            raise ValueError('There must be at least 2 DataFrames.')
        for i in args:
            if not isinstance(i, str):
                raise ValueError('Inputs must be Strings (names of DataFrames).')
        self.args = args

    def _repr_html_(self):
        return '\n'.join(self.template.format(i, eval(i)._repr_html_()) for i in self.args)

    def __repr__(self):
        return '\n\n'.join(i + '\n' + repr(eval(i)) for i in self.args)

    def shape(self):
        """ Return max of row and column no. """
        r = max([eval(i).shape[0] for i in self.args])
        c = max([eval(i).shape[1] for i in self.args])
        return r, c




import linearmodels as lm


def lighten_lm_res(res, mode=1):
    """
    Lighten linearmodels result for the case of large sample regression
    If mode=0, remove all data except _resid and _fitted
    If mode=1, keep dependent so that res.nobs and res.rsquared_adj still work
    """
    res._cov_estimator = None
    res._wresid = None
    if mode == 0:
        res.model = None
    elif mode == 1:
        if type(res.model) == lm.iv.model._OLS:
            res.model.exog = None
            res.model.endog = None
            res.model.weights = None
            res.model.instruments = None  # if mute this display still work but not sure result
            res.model._index = None
            res.model._y = None
            res.model._wy = None
            res.model._x = None
            res.model._wx = None
            res.model._z = None
            res.model._wz = None

        if type(res.model) == lm.iv.absorbing.AbsorbingLS:
            res.model._exog = None
            res.model._absorb = None
            res.model._absorb_inter = None
            res.model._weight_data = None
            res.model._absorbed_dependent = None
            res.model._absorbed_exog = None
            res.model._index = None
            res.model._original_index = None
            res.model._regressors = None

    else:
        raise ValueError("mode should be either 0 or 1")
    return res



"""
Functions that are useful for explore sqlite data

ref: https://stackoverflow.com/questions/305378/list-of-tables-db-schema-dump-etc-using-the-python-sqlite3-api
"""

import sqlite3
import pandas as pd


def read_data(file_path):
    """
    return (conn, c)
    """
    conn = sqlite3.connect(file_path)  # 'example.db'
    c = conn.cursor()
    return conn, c


def table_info(conn, c):
    '''
    prints out all of the columns of every table in db
    c : cursor object
    conn : database connection object
    '''
    tables = c.execute("SELECT name FROM sqlite_master WHERE type='table';").fetchall()
    for table_name in tables:
        table_name = table_name[0]  # tables is a list of single item tuples
        table = pd.read_sql_query(f"SELECT * from \"{table_name}\" LIMIT 0", conn)
        print(table_name)
        for col in table.columns:
            print('\t' + col)
        print()


def extract_tables(conn, c, table_names):
    '''
    extract either one (str) or mulitple (list) or all ([]) tables
    return either DataFrame or Dictionary
    '''
    if isinstance(table_names, str):
        table = pd.read_sql_query(f"SELECT * from \"{table_names}\"", conn)
        return table
    elif isinstance(table_names, list):
        if not table_names:
            tables = c.execute("SELECT name FROM sqlite_master WHERE type='table';").fetchall()
            table_names = [tn[0] for tn in tables]
        tables = {}
        for table_name in table_names:
            tables[table_name] = pd.read_sql_query(f"SELECT * from \"{table_name}\"", conn)
        return tables
    else:
        assert TypeError("table_names must be either str or list or []")

# def combine_tables(conn, c, id_col):
#     '''
#     combine tables according to same id column
#     '''
